//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.IO;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMessageInfo {
    /*
     * OpenRead()
     */
    public Stream OpenRead(CancellationToken cancellationToken)
    {
      return OpenFetchMessageBodyStream(ImapMessageFetchBodyOptions.Default,
                                        cancellationToken,
                                        null);
    }

#if NET_4_5
    public Stream OpenRead(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenFetchMessageBodyStream(ImapMessageFetchBodyOptions.Default,
                                        CancellationToken.None,
                                        progress);
    }

    public Stream OpenRead(CancellationToken cancellationToken,
                           IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenFetchMessageBodyStream(ImapMessageFetchBodyOptions.Default,
                                        cancellationToken,
                                        progress);
    }
#endif

    public Stream OpenRead(ImapMessageFetchBodyOptions options,
                           CancellationToken cancellationToken)
    {
      return OpenFetchMessageBodyStream(options,
                                        cancellationToken,
                                        null);
    }

#if NET_4_5
    public Stream OpenRead(ImapMessageFetchBodyOptions options,
                           IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenFetchMessageBodyStream(options,
                                        CancellationToken.None,
                                        progress);
    }

    public Stream OpenRead(ImapMessageFetchBodyOptions options,
                           CancellationToken cancellationToken,
                           IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenFetchMessageBodyStream(options,
                                        cancellationToken,
                                        progress);
    }
#endif

    /*
     * OpenReadAsync()
     */
    public Task<Stream> OpenReadAsync()
    {
      return OpenFetchMessageBodyStreamAsync(ImapMessageFetchBodyOptions.Default,
                                             CancellationToken.None,
                                             null,
                                             s => s);
    }

    public Task<Stream> OpenReadAsync(CancellationToken cancellationToken)
    {
      return OpenFetchMessageBodyStreamAsync(ImapMessageFetchBodyOptions.Default,
                                             cancellationToken,
                                             null,
                                             s => s);
    }

#if NET_4_5
    public Task<Stream> OpenReadAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenFetchMessageBodyStreamAsync(ImapMessageFetchBodyOptions.Default,
                                             CancellationToken.None,
                                             progress,
                                             s => s);
    }

    public Task<Stream> OpenReadAsync(CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenFetchMessageBodyStreamAsync(ImapMessageFetchBodyOptions.Default,
                                             cancellationToken,
                                             progress,
                                             s => s);
    }
#endif

    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options)
    {
      return OpenFetchMessageBodyStreamAsync(options,
                                             CancellationToken.None,
                                             null,
                                             s => s);
    }

    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken)
    {
      return OpenFetchMessageBodyStreamAsync(options,
                                             cancellationToken,
                                             null,
                                             s => s);
    }

#if NET_4_5
    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenFetchMessageBodyStreamAsync(options,
                                             CancellationToken.None,
                                             progress,
                                             s => s);
    }

    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenFetchMessageBodyStreamAsync(options,
                                             cancellationToken,
                                             progress,
                                             s => s);
    }
#endif

    /*
     * ReadAsAsync<TResult>(Converter<Stream>)
     */
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter)
    {
      return ReadAsAsync(ImapMessageFetchBodyOptions.Default,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return ReadAsAsync(ImapMessageFetchBodyOptions.Default,
                         converter,
                         cancellationToken,
                         null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsync(ImapMessageFetchBodyOptions.Default,
                         converter,
                         CancellationToken.None,
                         progress);
    }

    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsync(ImapMessageFetchBodyOptions.Default,
                         converter,
                         cancellationToken,
                         progress);
    }
#endif

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter)
    {
      return ReadAsAsync(options,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return ReadAsAsync(options,
                         converter,
                         cancellationToken,
                         null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsync(options,
                         converter,
                         CancellationToken.None,
                         progress);
    }
#endif

    // implementation of ReadAsAsync<TResult>(Converter<Stream>)
#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
#else
    private Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                               Converter<Stream, TResult> converter,
                                               CancellationToken cancellationToken,
                                               object progress)
#endif
    {
      if (converter == null)
        throw new ArgumentNullException("converter");

      return OpenFetchMessageBodyStreamAsync(options,
                                             cancellationToken,
                                             progress,
                                             delegate (Stream s) {
        using (s) {
          return converter(s);
        }
      });
    }

    /*
     * ReadAsAsync<TResult>(Func<Stream, ...>)
     */

    /*
     * ReadAllBytesAsync()
     */
    public Task<byte[]> ReadAllBytesAsync()
    {
      return ReadAllBytesAsync(ImapMessageFetchBodyOptions.Default,
                               CancellationToken.None,
                               null);
    }

    public Task<byte[]> ReadAllBytesAsync(CancellationToken cancellationToken)
    {
      return ReadAllBytesAsync(ImapMessageFetchBodyOptions.Default,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<byte[]> ReadAllBytesAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllBytesAsync(ImapMessageFetchBodyOptions.Default,
                               CancellationToken.None,
                               progress);
    }

    public Task<byte[]> ReadAllBytesAsync(CancellationToken cancellationToken,
                                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllBytesAsync(ImapMessageFetchBodyOptions.Default,
                               cancellationToken,
                               progress);
    }
#endif

    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options)
    {
      return ReadAllBytesAsync(options,
                               CancellationToken.None,
                               null);
    }

    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options,
                                          CancellationToken cancellationToken)
    {
      return ReadAllBytesAsync(options,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options,
                                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllBytesAsync(options,
                               CancellationToken.None,
                               progress);
    }
#endif

    // implementation of ReadAllBytesAsync()
#if NET_4_5
    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options,
                                          CancellationToken cancellationToken,
                                          IProgress<ImapMessageFetchProgressInfo> progress)
#else
    private Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options,
                                           CancellationToken cancellationToken,
                                           object progress)
#endif
    {
      return ReadAsAsync(options,
                         s => s.ReadToEnd(),
                         cancellationToken,
                         progress);
    }

    /*
     * WriteToAsync(Stream)
     */
    public Task WriteToAsync(Stream stream)
    {
      return WriteToAsync(stream,
                          ImapMessageFetchBodyOptions.Default,
                          CancellationToken.None,
                          null);
    }

    public Task WriteToAsync(Stream stream,
                             CancellationToken cancellationToken)
    {
      return WriteToAsync(stream,
                          ImapMessageFetchBodyOptions.Default,
                          cancellationToken,
                          null);
    }

#if NET_4_5
    public Task WriteToAsync(Stream stream,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsync(stream,
                          ImapMessageFetchBodyOptions.Default,
                          CancellationToken.None,
                          progress);
    }

    public Task WriteToAsync(Stream stream,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsync(stream,
                          ImapMessageFetchBodyOptions.Default,
                          cancellationToken,
                          progress);
    }
#endif

    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options)
    {
      return WriteToAsync(stream,
                          options,
                          CancellationToken.None,
                          null);
    }

    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken)
    {
      return WriteToAsync(stream,
                          options,
                          cancellationToken,
                          null);
    }

#if NET_4_5
    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsync(stream,
                          options,
                          CancellationToken.None,
                          progress);
    }
#endif

    // implementation of WriteToAsync(Stream)
#if NET_4_5
    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
#else
    private Task WriteToAsync(Stream stream,
                              ImapMessageFetchBodyOptions options,
                              CancellationToken cancellationToken,
                              object progress)
#endif
    {
      if (stream == null)
        throw new ArgumentNullException("stream");
      if (!stream.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream("stream");

      return ProcessOpenedMessageStreamAsync(options,
                                             cancellationToken,
                                             progress,
                                             s => CopyFetchMessageBodyStream(s, stream));
    }

    private Task ProcessOpenedMessageStreamAsync(ImapMessageFetchBodyOptions options,
                                                 CancellationToken cancellationToken,
#if NET_4_5
                                                 IProgress<ImapMessageFetchProgressInfo> progress,
#else
                                                 object progress,
#endif
                                                 Action<Stream> process)
    {
      return OpenFetchMessageBodyStreamAsync(options,
                                             cancellationToken,
                                             progress,
                                             s => s).ContinueWith(t => {
        using (var stream = t.Result) {
          process(stream);
        }
      });
    }

    /*
     * WriteToAsync(BinaryWriter)
     */
    public Task WriteToAsync(BinaryWriter writer)
    {
      return WriteToAsync(writer,
                          ImapMessageFetchBodyOptions.Default,
                          CancellationToken.None,
                          null);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             CancellationToken cancellationToken)
    {
      return WriteToAsync(writer,
                          ImapMessageFetchBodyOptions.Default,
                          cancellationToken,
                          null);
    }

#if NET_4_5
    public Task WriteToAsync(BinaryWriter writer,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsync(writer,
                          ImapMessageFetchBodyOptions.Default,
                          CancellationToken.None,
                          progress);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsync(writer,
                          ImapMessageFetchBodyOptions.Default,
                          cancellationToken,
                          progress);
    }
#endif

    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options)
    {
      return WriteToAsync(writer,
                          options,
                          CancellationToken.None,
                          null);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken)
    {
      return WriteToAsync(writer,
                          options,
                          cancellationToken,
                          null);
    }

#if NET_4_5
    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsync(writer,
                          options,
                          CancellationToken.None,
                          progress);
    }
#endif

    // implementation of WriteToAsync(BinaryWriter)
#if NET_4_5
    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
#else
    private Task WriteToAsync(BinaryWriter writer,
                              ImapMessageFetchBodyOptions options,
                              CancellationToken cancellationToken,
                              object progress)
#endif
    {
      if (writer == null)
        throw new ArgumentNullException("writer");

      return ProcessOpenedMessageStreamAsync(options,
                                             cancellationToken,
                                             progress,
                                             s => CopyFetchMessageBodyStream(s, writer));
    }

    /*
     * SaveAsync()
     */
    public Task SaveAsync(string path)
    {
      return SaveAsync(path,
                       ImapMessageFetchBodyOptions.Default,
                       CancellationToken.None,
                       null);
    }

    public Task SaveAsync(string path,
                          CancellationToken cancellationToken)
    {
      return SaveAsync(path,
                       ImapMessageFetchBodyOptions.Default,
                       cancellationToken,
                       null);
    }

#if NET_4_5
    public Task SaveAsync(string path,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return SaveAsync(path,
                       ImapMessageFetchBodyOptions.Default,
                       CancellationToken.None,
                       progress);
    }

    public Task SaveAsync(string path,
                          CancellationToken cancellationToken,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return SaveAsync(path,
                       ImapMessageFetchBodyOptions.Default,
                       cancellationToken,
                       progress);
    }
#endif

    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options)
    {
      return SaveAsync(path,
                       options,
                       CancellationToken.None,
                       null);
    }

    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options,
                          CancellationToken cancellationToken)
    {
      return SaveAsync(path,
                       options,
                       cancellationToken,
                       null);
    }

#if NET_4_5
    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return SaveAsync(path,
                       options,
                       CancellationToken.None,
                       progress);
    }
#endif

    // implementation of SaveAsync
#if NET_4_5
    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options,
                          CancellationToken cancellationToken,
                          IProgress<ImapMessageFetchProgressInfo> progress)
#else
    private Task SaveAsync(string path,
                           ImapMessageFetchBodyOptions options,
                           CancellationToken cancellationToken,
                           object progress)
#endif
    {
      if (path == null)
        throw new ArgumentNullException("path");
      if (path.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("path");

      return ProcessOpenedMessageStreamAsync(options,
                                             cancellationToken,
                                             progress,
                                             s => {
        using (var fileStream = new FileStream(path, FileMode.Create, FileAccess.Write)) {
          CopyFetchMessageBodyStream(s, fileStream);
        }
      });
    }

    /*
     * DownloadAsync()
     */
    public Task<Stream> DownloadAsync()
    {
      return DownloadAsync(ImapMessageFetchBodyOptions.Default,
                           CancellationToken.None,
                           null);
    }

    public Task<Stream> DownloadAsync(CancellationToken cancellationToken)
    {
      return DownloadAsync(ImapMessageFetchBodyOptions.Default,
                           cancellationToken,
                           null);
    }

#if NET_4_5
    public Task<Stream> DownloadAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return DownloadAsync(ImapMessageFetchBodyOptions.Default,
                           CancellationToken.None,
                           progress);
    }

    public Task<Stream> DownloadAsync(CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return DownloadAsync(ImapMessageFetchBodyOptions.Default,
                           cancellationToken,
                           progress);
    }
#endif

    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options)
    {
      return DownloadAsync(options,
                           CancellationToken.None,
                           null);
    }

    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken)
    {
      return DownloadAsync(options,
                           cancellationToken,
                           null);
    }

#if NET_4_5
    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return DownloadAsync(options,
                           CancellationToken.None,
                           progress);
    }
#endif

    // implementation of DownloadAsync
#if NET_4_5
    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
#else
    private Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options,
                                       CancellationToken cancellationToken,
                                       object progress)
#endif
    {
      return OpenFetchMessageBodyStreamAsync(options,
                                             cancellationToken,
                                             progress,
                                             DownloadCore);
    }

    /*
     * OpenTextAsync()
     */
    public Task<StreamReader> OpenTextAsync()
    {
      return OpenTextAsync(iso8859_1,
                           ImapMessageFetchBodyOptions.Default,
                           CancellationToken.None,
                           null);
    }

    public Task<StreamReader> OpenTextAsync(Encoding charset)
    {
      return OpenTextAsync(charset,
                           ImapMessageFetchBodyOptions.Default,
                           CancellationToken.None,
                           null);
    }

    public Task<StreamReader> OpenTextAsync(ImapMessageFetchBodyOptions options)
    {
      return OpenTextAsync(iso8859_1,
                           options,
                           CancellationToken.None,
                           null);
    }

    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options)
    {
      return OpenTextAsync(charset,
                           options,
                           CancellationToken.None,
                           null);
    }

    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken)
    {
      return OpenTextAsync(charset,
                           options,
                           cancellationToken,
                           null);
    }

#if NET_4_5
    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenTextAsync(charset,
                           options,
                           CancellationToken.None,
                           progress);
    }
#endif

    // implementation of OpenTextAsync
#if NET_4_5
    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
#else
    private Task<StreamReader> OpenTextAsync(Encoding charset,
                                             ImapMessageFetchBodyOptions options,
                                             CancellationToken cancellationToken,
                                             object progress)
#endif
    {
      if (charset == null)
        throw new ArgumentNullException("charset");

      return OpenFetchMessageBodyStreamAsync(options,
                                             cancellationToken,
                                             progress,
                                             s => new StreamReader(s, charset));
    }

    /*
     * ReadAsAsync<TResult>(Converter<StreamReader>)
     */
    public Task<TResult> ReadAsAsync<TResult>(Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsync(iso8859_1,
                         ImapMessageFetchBodyOptions.Default,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsync(charset,
                         ImapMessageFetchBodyOptions.Default,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsync(iso8859_1,
                         options,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsync(charset,
                         options,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return ReadAsAsync(charset,
                         options,
                         converter,
                         cancellationToken,
                         null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsync(charset,
                         options,
                         converter,
                         CancellationToken.None,
                         progress);
    }
#endif

    // implementation of ReadAsAsync<TResult>(Converter<StreamReader>)
#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
#else
    private Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                               ImapMessageFetchBodyOptions options,
                                               Converter<StreamReader, TResult> converter,
                                               CancellationToken cancellationToken,
                                               object progress)
#endif
    {
      if (charset == null)
        throw new ArgumentNullException("charset");
      if (converter == null)
        throw new ArgumentNullException("converter");

      return OpenFetchMessageBodyStreamAsync(options,
                                             cancellationToken,
                                             progress,
                                             delegate (Stream s) {
        using (var reader = new StreamReader(s, charset)) {
          return converter(reader);
        }
      });
    }

    /*
     * ReadLinesAsync()
     */
#if false
    // implementation of ReadLinesAsync()
    public Task<IEnumerable<string>> ReadLinesAsync(Encoding charset,
                                                    ImapMessageFetchBodyOptions options,
                                                    CancellationToken cancellationToken,
                                                    IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsync(charset,
                         options,
                         reader => reader.ReadLines(), // yield return does not work
                         cancellationToken,
                         progress);
    }
#endif

    /*
     * ReadAllLinesAsync()
     */
    public Task<string[]> ReadAllLinesAsync()
    {
      return ReadAllLinesAsync(iso8859_1,
                               ImapMessageFetchBodyOptions.Default,
                               CancellationToken.None,
                               null);
    }

    public Task<string[]> ReadAllLinesAsync(Encoding charset)
    {
      return ReadAllLinesAsync(charset,
                               ImapMessageFetchBodyOptions.Default,
                               CancellationToken.None,
                               null);
    }

    public Task<string[]> ReadAllLinesAsync(ImapMessageFetchBodyOptions options)
    {
      return ReadAllLinesAsync(iso8859_1,
                               options,
                               CancellationToken.None,
                               null);
    }

    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options)
    {
      return ReadAllLinesAsync(charset,
                               options,
                               CancellationToken.None,
                               null);
    }

    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken)
    {
      return ReadAllLinesAsync(charset,
                               options,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllLinesAsync(charset,
                               options,
                               CancellationToken.None,
                               progress);
    }
#endif

    // implementation of ReadAllLinesAsync()
#if NET_4_5
    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
#else
    private Task<string[]> ReadAllLinesAsync(Encoding charset,
                                             ImapMessageFetchBodyOptions options,
                                             CancellationToken cancellationToken,
                                             object progress)
#endif
    {
      return ReadAsAsync(charset,
                         options,
                         reader => reader.ReadAllLines(),
                         cancellationToken,
                         progress);
    }

    /*
     * ReadAllTextAsync()
     */
    public Task<string> ReadAllTextAsync()
    {
      return ReadAllTextAsync(iso8859_1,
                              ImapMessageFetchBodyOptions.Default,
                              CancellationToken.None,
                              null);
    }

    public Task<string> ReadAllTextAsync(Encoding charset)
    {
      return ReadAllTextAsync(charset,
                              ImapMessageFetchBodyOptions.Default,
                              CancellationToken.None,
                              null);
    }

    public Task<string> ReadAllTextAsync(ImapMessageFetchBodyOptions options)
    {
      return ReadAllTextAsync(iso8859_1,
                              options,
                              CancellationToken.None,
                              null);
    }

    public Task<string> ReadAllTextAsync(Encoding charset,
                                         ImapMessageFetchBodyOptions options)
    {
      return ReadAllTextAsync(charset,
                              options,
                              CancellationToken.None,
                              null);
    }

    public Task<string> ReadAllTextAsync(Encoding charset,
                                         ImapMessageFetchBodyOptions options,
                                         CancellationToken cancellationToken)
    {
      return ReadAllTextAsync(charset,
                              options,
                              cancellationToken,
                              null);
    }

#if NET_4_5
    public Task<string> ReadAllTextAsync(Encoding charset,
                                         ImapMessageFetchBodyOptions options,
                                         IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllTextAsync(charset,
                              options,
                              CancellationToken.None,
                              progress);
    }
#endif

    // implementation of ReadAllTextAsync()
#if NET_4_5
    public Task<string> ReadAllTextAsync(Encoding charset,
                                         ImapMessageFetchBodyOptions options,
                                         CancellationToken cancellationToken,
                                         IProgress<ImapMessageFetchProgressInfo> progress)
#else
    private Task<string> ReadAllTextAsync(Encoding charset,
                                          ImapMessageFetchBodyOptions options,
                                          CancellationToken cancellationToken,
                                          object progress)
#endif
    {
      return ReadAsAsync(charset,
                         options,
                         reader => reader.ReadToEnd(),
                         cancellationToken,
                         progress);
    }
  }
}
#endif
