using System;
using NUnit.Framework;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapSortCriteriaTests {
    [Test]
    public void TestCombine()
    {
      var c1 = ImapSortCriteria.Size;
      var c2 = ImapSortCriteria.Arrival;

      Assert.AreEqual("(SIZE ARRIVAL)", ImapSortCriteria.Combine(c1, c2).ToString());
      Assert.AreEqual("(SIZE ARRIVAL)", (c1 + c2).ToString());
      Assert.AreEqual("(SIZE ARRIVAL)", c1.CombineWith(c2).ToString());

      Assert.AreEqual("(ARRIVAL SIZE)", (c2 + c1).ToString());
    }

    [Test]
    public void TestOperatorAdd_1()
    {
      var criteria = ImapSortCriteria.Date + ImapSortCriteria.Subject;

      Assert.AreEqual("(DATE SUBJECT)", criteria.ToString());

      CollectionAssert.AreEquivalent(new[] {ImapCapability.Sort},
                                     (criteria as IImapExtension).RequiredCapabilities);

      criteria = ImapSortCriteria.Subject + ImapSortCriteria.Date;

      Assert.AreEqual("(SUBJECT DATE)", criteria.ToString());

      CollectionAssert.AreEquivalent(new[] {ImapCapability.Sort},
                                     (criteria as IImapExtension).RequiredCapabilities);
    }

    [Test]
    public void TestOperatorAdd_2()
    {
      var criteria = ImapSortCriteria.Date + ImapSortCriteria.DisplayFrom;

      Assert.AreEqual("(DATE DISPLAYFROM)", criteria.ToString());

      CollectionAssert.AreEquivalent(new[] {ImapCapability.Sort, ImapCapability.SortDisplay},
                                     (criteria as IImapExtension).RequiredCapabilities);

      criteria = ImapSortCriteria.DisplayFrom + ImapSortCriteria.Date;

      Assert.AreEqual("(DISPLAYFROM DATE)", criteria.ToString());

      CollectionAssert.AreEquivalent(new[] {ImapCapability.Sort, ImapCapability.SortDisplay},
                                     (criteria as IImapExtension).RequiredCapabilities);
    }
  }
}
