//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Text;

namespace Smdn.Security.Authentication.Sasl.Client {
  /*
   * The SASL XOAUTH2 Mechanism
   * https://developers.google.com/gmail/xoauth2_protocol
   */
  [SaslMechanism("XOAUTH2", false, false)]
  public class XOAuth2Mechanism : SaslClientMechanism {
    public override bool ClientFirst {
      get { return true; }
    }

    public string User {
      get { CheckDisposed(); return user; }
      set { CheckDisposed(); user = value; }
    }

    public string AccessToken {
      get { CheckDisposed(); return accessToken; }
      set { CheckDisposed(); accessToken = value; }
    }

    public string LastErrorResults {
      get { CheckDisposed(); return lastErrorResults; }
    }

    public XOAuth2Mechanism()
      : this(null, null)
    {
    }

    public XOAuth2Mechanism(string user, string accessToken)
    {
      this.user = user;
      this.accessToken = accessToken;
    }

    public override void Initialize()
    {
      base.Initialize();

      step = 0;
      lastErrorResults = null;
    }

    protected override SaslExchangeStatus Exchange(ByteString serverChallenge, out ByteString clientResponse)
    {
      clientResponse = null;

      switch (step++) {
        case 0: // initial client response
          if (user == null)
            return SaslExchangeStatus.Failed;
          if (accessToken == null)
            return SaslExchangeStatus.Failed;

          var builder = new ByteStringBuilder();

          builder.Append("user=");
          builder.Append(user);
          builder.Append("\x0001"); // ^A
          builder.Append("auth=Bearer ");
          builder.Append(accessToken);
          builder.Append("\x0001\x0001"); // ^A^A

          clientResponse = builder.ToByteString(false);

          return SaslExchangeStatus.Continuing;

        case 1: // error response
          lastErrorResults = serverChallenge.ToString();
          return SaslExchangeStatus.Failed;

        default: // unexpected server challenge
          return SaslExchangeStatus.Failed;
      }
    }

    private int step;
    private string user;
    private string accessToken;
    private string lastErrorResults;
  }
}
