// ドキュメントおよびサンプルは以下のページにあります
// http://smdn.jp/works/libs/Smdn.Net.Imap4.Client/docs/

using System;
using Smdn.Net.Imap4.Client;

class MainClass {
  public static void Main(string[] args)
  {
    ImapSslConnection.ServerCertificateValidationCallback += delegate {
      // 適切な証明書の検証を行うように書き換えてください
      return true;
    };

    using (var client = new ImapClient(new Uri("imap://user@imap.example.net/"))) {
      client.Profile.UseTlsIfAvailable = true;    // 可能ならSSL/TLSでの接続を行う
      client.Profile.AllowInsecureLogin = false;  // 平文での認証を許可しない
      client.Connect("pass");

      Console.WriteLine("connected!!");

      // すべてのメールボックスとメールボックスのステータスを取得して列挙
      foreach (var mailbox in client.GetMailboxes(ImapMailboxListOptions.RequestStatus)) {
        // 合計メッセージ数と名前を表示
        Console.WriteLine("{0} ({1})", mailbox.FullName, mailbox.ExistMessageCount);
      }

      // INBOXを開く
      using (var inbox = client.OpenInbox()) {
        // 合計・未読・新着メッセージ数
        Console.WriteLine("[INBOX]");
        Console.WriteLine("total message count : {0}", inbox.ExistMessageCount);
        Console.WriteLine("unseen message count : {0}", inbox.UnseenMessageCount);
        Console.WriteLine("recent message count : {0}", inbox.RecentMessageCount);

        // すべてのメッセージとその属性を取得
        foreach (var message in inbox.GetMessages(ImapMessageFetchAttributeOptions.AllAttributes)) {
          if (message.IsRecent)
            Console.Write("!"); // 新着
          else if (!message.IsSeen)
            Console.Write("*"); // 未読
          else
            Console.Write(" ");

          // 着信日時とタイトルを表示
          Console.WriteLine(" ({0:o}) {1}", message.InternalDate, message.EnvelopeSubject);
        }
      }
    }
  }
}